from __future__ import annotations

import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable

from visisipy.backend import get_backend

if TYPE_CHECKING:
    from visisipy.models import EyeModel


def _validate_analysis_signature(function: Callable[..., tuple[Any, Any]]) -> None:
    """
    Validate the signature of an analysis function.

    The first parameter must be 'model' with type 'EyeModel | None'. The last parameter must be 'return_raw_result'
    with type 'bool'.

    Parameters
    ----------
    function : Callable
        The analysis function to validate.

    Raises
    ------
    ValueError
        If the signature of the analysis function is invalid.
    """

    signature = inspect.signature(function)
    first_parameter, *_, last_parameter = signature.parameters

    if first_parameter != "model":
        raise ValueError("The first parameter of an analysis function must be 'model'.")

    if signature.parameters["model"].annotation != "EyeModel | None":
        raise ValueError(
            f"The first parameter of an analysis function must have type 'EyeModel | None', "
            f"got '{signature.parameters['model'].annotation}'"
        )

    if last_parameter != "return_raw_result":
        raise ValueError("The last parameter of an analysis function must be 'return_raw_result'.")

    if signature.parameters["return_raw_result"].annotation != "bool":
        raise ValueError(
            f"The last parameter of an analysis function must have type 'bool', "
            f"got '{signature.parameters['return_raw_result'].annotation}'"
        )


def _build_model(model: EyeModel) -> None:
    """
    Build the model in OpticStudio if it is not already built.

    Parameters
    ----------
    model : Any
        The model to build.
    """
    backend = get_backend()

    if backend.model is None or backend.model.eye_model is not model:
        backend.build_model(model)


def analysis(function: Callable[..., tuple[Any, Any]]) -> Callable:
    """
    Decorator for analysis functions.

    This decorator is used to mark a function as an analysis function. Analysis functions are used to perform various
    analyses on the optical system.

    Parameters
    ----------
    function : Callable
        The analysis function.

    Returns
    -------
    Callable
        The analysis function.
    """
    _validate_analysis_signature(function)

    @wraps(function)
    def analysis_wrapper(
        model: EyeModel | None = None,
        *args: Any,
        return_raw_result: bool = False,
        **kwargs: Any,
    ) -> tuple[Any, Any]:
        if model is not None:
            _build_model(model)

        result, raw_result = function(model, *args, return_raw_result=return_raw_result, **kwargs)

        if return_raw_result:
            return result, raw_result

        return result

    return analysis_wrapper



from __future__ import annotations

from visisipy.opticstudio.backend import initialize_opticstudio
from visisipy.opticstudio.models import OpticStudioEye

__all__ = ("OpticStudioEye", "initialize_opticstudio")



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--no-opticstudio",
        action="store_true",
        help="Skip tests that require OpticStudio.",
    )


def pytest_collection_modifyitems(config, items):
    for item in items:
        needs_opticstudio = item.get_closest_marker("needs_opticstudio")
        if needs_opticstudio and config.getoption("--no-opticstudio"):
            item.add_marker("skip")


def detect_opticstudio() -> bool:
    import zospy as zp

    opticstudio_available: bool = False

    try:
        zos = zp.ZOS()
        opticstudio_available = True

        zos.disconnect()
        assert zos.Application is None

        zp.ZOS._instances = set()  # noqa: SLF001
        del zos

    except FileNotFoundError:
        opticstudio_available = False

    del zp
    return opticstudio_available


@pytest.fixture(scope="session")
def opticstudio_available(request) -> bool:
    if request.config.getoption("--no-opticstudio"):
        return False

    return detect_opticstudio()


@pytest.fixture(autouse=True)
def skip_opticstudio(request, opticstudio_available):
    if request.node.get_closest_marker("needs_opticstudio") and not opticstudio_available:
        pytest.skip("OpticStudio is not available.")



from contextlib import nullcontext as does_not_raise

import pytest

from visisipy.wavefront import ZernikeCoefficients


class TestZernikeCoefficients:
    @pytest.mark.parametrize(
        "terms,result,expectation",
        [
            (None, {}, does_not_raise()),
            ({}, {}, does_not_raise()),
            ({1: 0.5, 2: 0.9}, {1: 0.5, 2: 0.9}, does_not_raise()),
            ({1: 0.5, "x": 0.9}, {}, pytest.raises(TypeError, match="All keys must be integers")),
            ({1: 0.5, 2.8: 0.9}, {}, pytest.raises(TypeError, match="All keys must be integers")),
            ({1: 0.5, 0: 0.9}, {}, pytest.raises(ValueError, match="The Zernike coefficients must be larger than 0")),
            ({-1: 0.5, 2: 0.9}, {}, pytest.raises(ValueError, match="The Zernike coefficients must be larger than 0")),
        ],
    )
    def test_init(self, terms, result, expectation):
        with expectation:
            zernike = ZernikeCoefficients(terms)
            assert zernike == result

    @pytest.mark.parametrize(
        "n,value,expectation",
        [
            (1, 1.2, does_not_raise()),
            (1, 0.0, does_not_raise()),
            (0, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            (-1, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            ("x", 0.0, pytest.raises(TypeError, match="The key must be an integer")),
            (1.5, 0.0, pytest.raises(TypeError, match="The key must be an integer")),
        ],
    )
    def test_setitem(self, n: int, value: float, expectation):
        zernike = ZernikeCoefficients()

        with expectation:
            zernike[n] = value
            assert zernike[n] == value

    @pytest.mark.parametrize(
        "n,expected_value,expectation",
        [
            (1, 0.5, does_not_raise()),
            (2, 0.9, does_not_raise()),
            (3, 0.0, does_not_raise()),
            (0, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            (-1, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            ("x", 0.0, pytest.raises(TypeError, match="The key must be an integer")),
            (1.5, 0.0, pytest.raises(TypeError, match="The key must be an integer")),
        ],
    )
    def test_getitem(self, n, expected_value, expectation):
        zernike = ZernikeCoefficients({1: 0.5, 2: 0.9})

        with expectation:
            assert zernike[n] == expected_value



from __future__ import annotations

import pytest
import zospy as zp

from visisipy.opticstudio.backend import OpticStudioBackend


@pytest.fixture(scope="session")
def monkeypatch_session():
    """
    Session-scoped monkeypatch fixture.
    """
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="session")
def zos(monkeypatch_session):
    """
    Initialize a ZOS instance.

    After initializing the ZOS instance, zospy.ZOS is patched to return the same instance,
    so multiple calls to zospy.ZOS() will not raise an error.
    """
    zos = zp.ZOS()

    def patched_zos(
        preload: bool = False,  # noqa: FBT001, FBT002
        zosapi_nethelper: str | None = None,
        opticstudio_directory: str | None = None,
    ):
        return zos

    monkeypatch_session.setattr(zp, "ZOS", patched_zos)

    yield zos

    zos.disconnect()


@pytest.fixture(scope="session")
def oss(zos):
    oss = zos.connect(mode="standalone")

    yield oss

    oss.close()


@pytest.fixture
def new_oss(oss):
    oss.new()

    return oss


@pytest.fixture
def opticstudio_backend(zos, monkeypatch):
    OpticStudioBackend.initialize(mode="standalone")

    yield OpticStudioBackend

    if OpticStudioBackend.zos is not None and OpticStudioBackend.oss is not None:
        OpticStudioBackend.disconnect()



from __future__ import annotations

from types import SimpleNamespace

import pytest

from visisipy import EyeModel
from visisipy.analysis import base


@pytest.fixture
def mock_backend(monkeypatch):
    """Mock the backend for testing."""

    class MockBackend:
        def __init__(self):
            self.model = None

        def build_model(self, model):
            self.model = SimpleNamespace(eye_model=model)

    backend = MockBackend()

    monkeypatch.setattr(base, "get_backend", lambda: backend)

    return backend


@pytest.fixture
def example_analysis():
    """Example analysis function to test the analysis decorator."""

    def example_analysis(model: EyeModel | None, x: int, *, return_raw_result: bool):
        return x, x

    return example_analysis


class TestAnalysisDecorator:
    def test_decorator(self, mock_backend, example_analysis):
        decorated_analysis = base.analysis(example_analysis)

        assert decorated_analysis(EyeModel(), 1) == 1
        assert decorated_analysis(EyeModel(), 1, return_raw_result=True) == (1, 1)

    def test_build_model(self, mock_backend, example_analysis):
        model = EyeModel()

        decorated_analysis = base.analysis(example_analysis)
        decorated_analysis(model, 1)

        assert mock_backend.model is not None
        assert mock_backend.model.eye_model == model

    def test_no_model_raises_valueerror(self):
        def example_analysis(x, *, return_raw_result: bool = False):
            return x, x

        with pytest.raises(
            ValueError,
            match="The first parameter of an analysis function must be 'model'",
        ):
            base.analysis(example_analysis)

    def test_invalid_model_annotation_raises_valueerror(self):
        def example_analysis(model: str, x, *, return_raw_result: bool = False):
            return x, x

        with pytest.raises(
            ValueError,
            match="The first parameter of an analysis function must have type 'EyeModel | None', got 'str'",
        ):
            base.analysis(example_analysis)

    def test_no_return_raw_result_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x):
            return x, x

        with pytest.raises(
            ValueError,
            match="The last parameter of an analysis function must be 'return_raw_result'",
        ):
            base.analysis(example_analysis)

    def test_invalid_return_raw_result_annotation_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x, return_raw_result: str):
            return x, x

        with pytest.raises(
            ValueError,
            match="The last parameter of an analysis function must have type 'bool', got 'str'",
        ):
            base.analysis(example_analysis)



"""Utilities for handling wavefront data."""

from __future__ import annotations

from collections import defaultdict

__all__ = ("ZernikeCoefficients",)


class ZernikeCoefficients(defaultdict):
    """Zernike coefficients.

    Convenience class for handling Zernike coefficients as a dictionary. If a term is not present, 0 is returned.
    Upon initialization and setting items, the keys are validated to be non-negative integers.

    Raises
    ------
    TypeError
        If the key is not an integer.
    ValueError
        If the key is smaller than 1.
    """

    def __init__(self, terms: dict[int, float] | None = None):
        if terms is not None:
            if not all(isinstance(key, int) for key in terms):
                raise TypeError("All keys must be integers.")
            if any(key < 1 for key in terms):
                raise ValueError("The Zernike coefficients must be larger than 0.")

        super().__init__(lambda: 0, terms or {})

    @staticmethod
    def _validate_coefficient(key: int) -> None:
        if not isinstance(key, int):
            raise TypeError("The key must be an integer.")
        if key < 1:
            raise ValueError("The coefficient must be larger than 0.")

    def __setitem__(self, key: int, value: float) -> None:
        self._validate_coefficient(key)

        super().__setitem__(key, value)

    def __getitem__(self, key: int) -> float:
        self._validate_coefficient(key)

        return super().__getitem__(key)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self)})"



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import visisipy.backend as _backend
from visisipy.models.geometry import EyeGeometry, NavarroGeometry, create_geometry
from visisipy.models.materials import EyeMaterials, NavarroMaterials

if TYPE_CHECKING:
    from os import PathLike

__all__ = (
    "create_geometry",
    "EyeModel",
    "EyeGeometry",
    "NavarroGeometry",
    "EyeMaterials",
    "NavarroMaterials",
    "BaseEye",
    "BaseSurface",
)


@dataclass
class EyeModel:
    geometry: EyeGeometry = field(default_factory=NavarroGeometry)
    materials: EyeMaterials = field(default_factory=NavarroMaterials)
    _built: BaseEye | None = field(default=None, init=False, repr=False)

    def build(self, **kwargs) -> BaseEye:
        backend = _backend.get_backend()
        self._built = backend.build_model(self, **kwargs)

        return self._built

    @staticmethod
    def clear() -> None:
        backend = _backend.get_backend()
        backend.clear_model()

    @staticmethod
    def save(filename: str | PathLike | None = None) -> None:
        backend = _backend.get_backend()
        backend.save_model(filename)


class BaseSurface(ABC):
    """Abstract class that must be implemented by backend-specific surface classes."""

    @abstractmethod
    def __init__(
        self,
        comment: str,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: Any | None = None,
        *,
        is_stop: bool | None = None,
    ): ...

    @property
    @abstractmethod
    def surface(self) -> Any: ...

    @abstractmethod
    def build(self, *args, **kwargs):
        """Build the surface in the backend."""
        ...


class BaseEye(ABC):
    """Abstract class that must be implemented by backend-specific eye model classes."""

    @abstractmethod
    def __init__(self, model: EyeModel): ...

    @abstractmethod
    def build(self, *args, **kwargs):
        """Build the eye model in the backend."""
        ...

    @property
    @abstractmethod
    def eye_model(self) -> EyeModel: ...

    @property
    def surfaces(self) -> dict[str, BaseSurface]:
        """Dictionary with surface names as keys and surfaces as values."""
        return {k.lstrip("_"): v for k, v in self.__dict__.items() if isinstance(v, BaseSurface)}

    def update_surfaces(self, attribute: str, value: Any, surfaces: list[str] | None = None) -> None:
        """Batch update all surfaces.

        Set `attribute` to `value` for multiple surfaces. If `surfaces` is not specified, all surfaces of the eye
        model are updated.

        Parameters
        ----------
        attribute : str
            Name of the attribute to update
        value : Any
            New value of the surface attribute
        surfaces : list[str]
            List of surfaces to be updated. If not specified, all surfaces are updated.

        Returns
        -------

        """
        surfaces = [self.surfaces[s] for s in surfaces] if surfaces is not None else self.surfaces.keys()

        for s in surfaces:
            setattr(s, attribute, value)



from __future__ import annotations

import importlib
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Literal
from warnings import warn

if TYPE_CHECKING:
    from collections.abc import Iterable
    from os import PathLike

    from pandas import DataFrame
    from zospy.zpcore import OpticStudioSystem

    from visisipy.models import BaseEye, EyeModel
    from visisipy.refraction import FourierPowerVectorRefraction

_BACKEND: BaseBackend | None = None
_DEFAULT_BACKEND: Backend | str = "opticstudio"


class _classproperty(property):  # noqa: N801
    def __get__(self, instance, owner=None):
        return self.fget(owner)


class BaseAnalysis(ABC):
    @staticmethod
    @abstractmethod
    def raytrace(
        self,
        coordinates: Iterable[tuple[float, float]],
        field_type: Literal["angle", "object"] = "angle",
        pupil: tuple[float, float] = (0, 0),
    ) -> DataFrame: ...

    @staticmethod
    @abstractmethod
    def refraction(
        self,
        field_coordinate: tuple[float, float] | None = None,
        wavelength: float | None = None,
    ) -> FourierPowerVectorRefraction: ...


class BaseBackend(ABC):
    model: BaseEye | None

    @_classproperty
    def analysis(self) -> BaseAnalysis: ...

    @classmethod
    @abstractmethod
    def build_model(cls, model: EyeModel, **kwargs) -> BaseEye: ...

    @classmethod
    @abstractmethod
    def clear_model(cls) -> None: ...

    @classmethod
    @abstractmethod
    def save_model(cls, filename: str | PathLike | None = None) -> None: ...


class Backend(str, Enum):
    OPTICSTUDIO = "opticstudio"


def set_backend(backend: Backend | str = Backend.OPTICSTUDIO, **kwargs) -> None:
    """Set the backend to use for optical simulations.

    Parameters
    ----------
    backend : Backend | str
        The backend to use. Must be one of the values in the `Backend` enum. Defaults to `Backend.OPTICSTUDIO`.
    kwargs
        Additional keyword arguments to pass to the backend's `initialize` method.

    Raises
    ------
    ValueError
        If an invalid backend is specified.
    """
    global _BACKEND  # noqa: PLW0603

    if _BACKEND is not None:
        warn(
            f"The backend is already set to {_BACKEND.__class__.__name__}. "
            f"Reconfiguring the backend is not recommended and may cause issues."
        )

    if backend == Backend.OPTICSTUDIO:
        os_backend = importlib.import_module("visisipy.opticstudio.backend")
        _BACKEND = os_backend.OpticStudioBackend
        _BACKEND.initialize(**kwargs)
    else:
        raise ValueError(f"Unknown backend: {backend}")


def get_backend() -> BaseBackend:
    """Get the current backend.

    Returns
    -------
    BaseBackend
        The current backend.
    """
    if _BACKEND is None:
        set_backend(_DEFAULT_BACKEND)

    return _BACKEND


def get_oss() -> OpticStudioSystem | None:
    """
    Get the OpticStudioSystem instance from the current backend.

    Returns
    -------
    OpticStudioSystem | None
        The OpticStudioSystem instance if the current backend is the OpticStudio backend, otherwise `None`.
    """
    os_backend = importlib.import_module("visisipy.opticstudio.backend")

    if _BACKEND is os_backend.OpicStudioBackend:
        return os_backend._oss  # noqa: SLF001

    return None



from __future__ import annotations

from visisipy import analysis, models, opticstudio, plots, refraction, wavefront
from visisipy.backend import get_backend, set_backend
from visisipy.models import (
    EyeGeometry,
    EyeMaterials,
    EyeModel,
    NavarroGeometry,
    NavarroMaterials,
)

__all__ = (
    "models",
    "EyeModel",
    "EyeGeometry",
    "EyeMaterials",
    "NavarroGeometry",
    "NavarroMaterials",
    "analysis",
    "opticstudio",
    "get_backend",
    "set_backend",
    "plots",
    "refraction",
    "wavefront",
)

__version__ = "0.0.1"


from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from typing import Any, NamedTuple, TypeVar
from warnings import warn

import numpy as np

from visisipy.wavefront import ZernikeCoefficients

__all__ = (
    "Surface",
    "StandardSurface",
    "Stop",
    "EyeGeometry",
    "NavarroGeometry",
    "create_geometry",
)


@dataclass
class Surface(ABC):  # noqa: B024
    """Base class for optical surfaces.

    Attributes
    ----------
    thickness : float
        The thickness of the surface. Default is 0.
    """

    thickness: float = 0


class _HalfAxes2D(NamedTuple):
    axial: float
    radial: float


@dataclass
class StandardSurface(Surface):
    """Standard conic surface.

    Represents surfaces as conic sections (spheres, ellipsoids, paraboloids, hyperboloids), defined in terms of their
    radius of curvature and asphericity. For the asphericity, the following definition is used:

    .. math::
        k = - \varepsilon^2 = - \\left( 1 - \frac{b^2}{a^2} \right)

    with :math:`k` the asphericity, :math:`\varepsilon` the eccentricity, :math:`a` the ellipsoid axis parallel to
    the optical axis and :math:`b` the ellipsoid axis perpendicular to the optical axis. This is the definition used
    in OpticStudio. Note that the meaning of :math:`a` and :math:`b` differs from the standard definition of an ellipse,
    where they are the semi-major and semi-minor axes.

    Attributes
    ----------
    radius : float
        The radius of the surface. Default is infinity.
    asphericity : float
        The asphericity of the surface. Default is 0.
    thickness : float
        The thickness of the surface. Default is 0.
    semi_diameter : float | None
        The semi-diameter of the surface aperture. Default is `None`.

    Methods
    -------
    half_axes(self) -> tuple[float, float]:
        Calculates and returns the half axes of the surface.
    """

    radius: float = float("inf")
    asphericity: float = 0
    thickness: float = 0
    semi_diameter: float | None = None
    is_stop: bool = False

    @property
    def half_axes(self) -> _HalfAxes2D:
        """Calculates and returns the half axes of the surface.

        This works only if the surface is an ellipsoid (asphericity > -1), otherwise a NotImplementedError is raised.
        A tuple of the axial and radial half axes is returned, with the axial axis being parallel to the optical axis
        and the radial axis perpendicular to it.

        Returns
        -------
        tuple[float, float]
            The axial and radial half axes of the surface.

        Raises
        ------
        NotImplementedError
            If the surface is not an ellipsoid (asphericity <= -1).
        """
        if self.asphericity > -1:
            return _HalfAxes2D(
                axial=(self.radius / (self.asphericity + 1)),
                radial=abs(self.radius / np.sqrt(self.asphericity + 1)),
            )

        raise NotImplementedError(
            f"Half axes are only defined for ellipses (asphericity > -1), " f"got {self.asphericity=}"
        )


@dataclass
class Stop(StandardSurface):
    """Stop surface.

    This surface represents the aperture stop of an optical system.

    Attributes
    ----------
    semi_diameter : float
        The semi-diameter of the aperture stop. Default is 1.
    thickness : float
        The thickness of the stop. Default is 0.
    """

    semi_diameter: float = 1.0
    thickness: float = 0
    is_stop: bool = True


@dataclass
class ZernikeStandardSagSurface(StandardSurface):
    """Zernike standard coefficients surface with surface deformations.

    Represents a surface with surface deformations described by Zernike polynomials.

    Attributes
    ----------
    zernike_coefficients : ZernikeCoefficients | dict[int, float]
        The Zernike coefficients of the surface. Default is an empty dictionary.
    extrapolate : bool
        If True, the Zernike coefficients will be considered even if the ray lands beyond the normalization radius.
        Default is `True`.
    zernike_decenter_x : float
        Decentration of the Zernike terms with respect to the conical and aspherical terms in the x-direction.
        Default is 0.
    zernike_decenter_y : float
        Decentration of the Zernike terms with respect to the conical and aspherical terms in the y-direction.
        Default is 0.
    maximum_term : int
        The maximum Zernike term to consider. Default is 0.
    norm_radius : float
        The normalization radius for the Zernike coefficients, in lens units (usually mm). Default is 100.

    Raises
    ------
    ValueError
        If the Zernike coefficients contain terms that are greater than the maximum term.
    """

    zernike_coefficients: ZernikeCoefficients | dict[int, float] = field(default_factory=dict)
    extrapolate: bool = True
    zernike_decenter_x: float = 0
    zernike_decenter_y: float = 0
    maximum_term: int = 0
    norm_radius: float = 100

    def __post_init__(self):
        if any(key > self.maximum_term for key in self.zernike_coefficients):
            raise ValueError("The Zernike coefficients contain terms that are greater than the maximum term.")

        self.zernike_coefficients = ZernikeCoefficients(self.zernike_coefficients)


@dataclass
class ZernikeStandardPhaseSurface(StandardSurface):
    """Zernike standard coefficients surface with wavefront aberrations.

    Represents a surface with wavefront aberrations described by Zernike polynomials.

    Attributes
    ----------
    zernike_coefficients : ZernikeCoefficients | dict[int, float]
        The Zernike coefficients of the surface. Default is an empty dictionary.
    extrapolate : bool
        If True, the Zernike coefficients will be considered even if the ray lands beyond the normalization radius.
        Default is `True`.
    diffraction_order : float
        The diffraction order of the surface. Default is 0.
    maximum_term : int
        The maximum Zernike term to consider. Default is 0.
    norm_radius : float
        The normalization radius for the Zernike coefficients, in lens units (usually mm). Default is 100.

    Raises
    ------
    ValueError
        If the Zernike coefficients contain terms that are greater than the maximum term.
    """

    zernike_coefficients: ZernikeCoefficients | dict[int, float] = field(default_factory=dict)
    extrapolate: bool = True
    diffraction_order: float = 1
    maximum_term: int = 0
    norm_radius: float = 100

    def __post_init__(self):
        if any(key > self.maximum_term for key in self.zernike_coefficients):
            raise ValueError("The Zernike coefficients contain terms that are greater than the maximum term.")

        self.zernike_coefficients = ZernikeCoefficients(self.zernike_coefficients)


class EyeGeometry:
    """Geometric parameters of an eye.

    Sizes are specified in mm. This class is mainly intended as a base class for more specific eye models.
    When used directly, all surfaces need to be specified manually.

    Attributes
    ----------
    cornea_front : Surface
        The front surface of the cornea.
    cornea_back : Surface
        The back surface of the cornea.
    pupil : Surface
        The pupil of the eye.
    lens_front : Surface
        The front surface of the lens.
    lens_back : Surface
        The back surface of the lens.
    retina : Surface
        The retina of the eye.

    Methods
    -------
    axial_length(self) -> float:
        Calculates and returns the axial length of the eye.
    cornea_thickness(self) -> float:
        Returns the thickness of the cornea.
    anterior_chamber_depth(self) -> float:
        Returns the depth of the anterior chamber.
    lens_thickness(self) -> float:
        Returns the thickness of the lens.
    vitreous_thickness(self) -> float:
        Returns the thickness of the vitreous.
    """

    def __init__(
        self,
        cornea_front: StandardSurface | None = None,
        cornea_back: StandardSurface | None = None,
        pupil: Stop | None = None,
        lens_front: StandardSurface | None = None,
        lens_back: StandardSurface | None = None,
        retina: StandardSurface | None = None,
    ):
        """Initializes an instance of the EyeGeometry class.

        This creates an instance of the EyeGeometry class with the specified surfaces. If a surface is not provided,
        the default values of `StandardSurface` and `Stop` will be used. This class is mainly intended as a base
        class, and instantiating it without specifying the surfaces will not result in a realistic eye model. To
        initialize a realistic eye model, use a subclass such as `NavarroGeometry`.

        Parameters
        ----------
        cornea_front : Surface, optional
            The front surface of the cornea. If not provided, a default Surface instance will be used.
        cornea_back : Surface, optional
            The back surface of the cornea. If not provided, a default Surface instance will be used.
        pupil : Surface, optional
            The pupil of the eye. If not provided, a default Stop instance will be used.
        lens_front : Surface, optional
            The front surface of the lens. If not provided, a default Surface instance will be used.
        lens_back : Surface, optional
            The back surface of the lens. If not provided, a default Surface instance will be used.
        retina : StandardSurface, optional
            The retina of the eye. If not provided, a default Surface instance will be used.
        """

        self.cornea_front = cornea_front or StandardSurface()
        self.cornea_back = cornea_back or StandardSurface()
        self.pupil = pupil or Stop()
        self.lens_front = lens_front or StandardSurface()
        self.lens_back = lens_back or StandardSurface()
        self.retina = retina or StandardSurface()

    @property
    def cornea_front(self) -> StandardSurface:
        """Cornea front geometry."""
        return self._cornea_front

    @cornea_front.setter
    def cornea_front(self, value: StandardSurface) -> None:
        self._cornea_front = value

    @property
    def cornea_back(self) -> StandardSurface:
        """Cornea back geometry."""
        return self._cornea_back

    @cornea_back.setter
    def cornea_back(self, value: StandardSurface) -> None:
        self._cornea_back = value

    @property
    def pupil(self) -> Stop:
        """Pupil geometry."""
        return self._pupil

    @pupil.setter
    def pupil(self, value: Stop) -> None:
        if not value.is_stop:
            raise ValueError("The pupil surface must be a stop.")

        self._pupil = value

    @property
    def lens_front(self) -> StandardSurface:
        """Lens front geometry."""
        return self._lens_front

    @lens_front.setter
    def lens_front(self, value: StandardSurface) -> None:
        self._lens_front = value

    @property
    def lens_back(self) -> StandardSurface:
        """Lens back geometry."""
        return self._lens_back

    @lens_back.setter
    def lens_back(self, value: StandardSurface) -> None:
        self._lens_back = value

    @property
    def retina(self) -> StandardSurface:
        """Retina geometry."""
        return self._retina

    @retina.setter
    def retina(self, value: StandardSurface) -> None:
        if value.asphericity <= -1:
            raise ValueError(f"Only an elliptical retina is allowed (asphericity > -1), got {value.asphericity=}")

        self._retina = value

    @property
    def axial_length(self) -> float:
        """Axial length of the eye, in mm."""
        return (
            self.cornea_front.thickness
            + self.cornea_back.thickness
            + self.lens_front.thickness
            + self.lens_back.thickness
        )

    @property
    def cornea_thickness(self) -> float:
        """Thickness of the cornea, in mm."""
        return self.cornea_front.thickness

    @property
    def anterior_chamber_depth(self) -> float:
        """Depth of the anterior chamber, in mm."""
        return self.cornea_back.thickness

    @property
    def lens_thickness(self) -> float:
        """Thickness of the crystalline lens, in mm."""
        return self.lens_front.thickness

    @property
    def vitreous_thickness(self) -> float:
        """Thickness of the vitreous body, in mm."""
        return self.lens_back.thickness

    def reverse(self):
        raise NotImplementedError


class NavarroGeometry(EyeGeometry):
    """Geometric parameters of the Navarro schematic eye.

    This schematic eye is based on the Navarro model as described in [1]_.
    Sizes are specified in mm.

     .. [1] Escudero-Sanz, I., & Navarro, R. (1999). Off-axis aberrations of a wide-angle schematic eye model.
            JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881

    Attributes
    ----------
    cornea_front : StandardSurface
        The front surface of the cornea.
    cornea_back : StandardSurface
        The back surface of the cornea.
    pupil : Stop
        The pupil of the eye.
    lens_front : StandardSurface
        The front surface of the lens.
    lens_back : StandardSurface
        The back surface of the lens.
    retina : StandardSurface
        The retina of the eye.

    Examples
    --------
    Use the default Navarro geometry:

    >>> from visisipy import NavarroGeometry
    >>> geometry = NavarroGeometry()

    Create a Navarro geometry with a custom retina:

    >>> geometry = NavarroGeometry(
    ...     retina=StandardSurface(radius=-12.5, asphericity=0.5)
    ... )

    Create a default Navarro gometry and change only the lens back radius:

    >>> geometry = NavarroGeometry()
    >>> geometry.lens_back.radius = -5.8
    """

    def __init__(self, **kwargs):
        surfaces = {
            "cornea_front": StandardSurface(radius=7.72, asphericity=-0.26, thickness=0.55),
            "cornea_back": StandardSurface(radius=6.50, asphericity=0, thickness=3.05),
            "pupil": Stop(semi_diameter=1.348),
            "lens_front": StandardSurface(radius=10.2, asphericity=-3.1316, thickness=4.0),
            "lens_back": StandardSurface(radius=-6.0, asphericity=-1, thickness=16.3203),
            "retina": StandardSurface(radius=-12.0, asphericity=0),
        }
        surfaces.update(**kwargs)
        super().__init__(**surfaces)


def _update_attribute_if_specified(obj: Surface, attribute: str, value: Any):
    if value is not None:
        setattr(obj, attribute, value)


def _calculate_vitreous_thickness(
    geometry: EyeGeometry,
    axial_length: float | None = None,
    cornea_thickness: float | None = None,
    anterior_chamber_depth: float | None = None,
    lens_thickness: float | None = None,
) -> float:
    """Calculate the thickness of the vitreous body for a partially initialized eye geometry."""
    _axial_length = geometry.axial_length if axial_length is None else axial_length
    _cornea_thickness = geometry.cornea_thickness if cornea_thickness is None else cornea_thickness
    _anterior_chamber_depth = (
        geometry.anterior_chamber_depth if anterior_chamber_depth is None else anterior_chamber_depth
    )
    _lens_thickness = geometry.lens_thickness if lens_thickness is None else lens_thickness

    if None in (
        _axial_length,
        _cornea_thickness,
        _anterior_chamber_depth,
        _lens_thickness,
    ):
        raise ValueError("Cannot calculate vitreous thickness from the supplied parameters.")

    return _axial_length - (_cornea_thickness + _anterior_chamber_depth + _lens_thickness)


GeometryType = TypeVar("GeometryType", bound=EyeGeometry)


def create_geometry(
    base: type[GeometryType] = NavarroGeometry,
    axial_length: float | None = None,
    cornea_thickness: float | None = None,
    cornea_front_radius: float | None = None,
    cornea_front_asphericity: float | None = None,
    cornea_back_radius: float | None = None,
    cornea_back_asphericity: float | None = None,
    anterior_chamber_depth: float | None = None,
    pupil_radius: float | None = None,
    lens_thickness: float | None = None,
    lens_back_radius: float | None = None,
    lens_back_asphericity: float | None = None,
    lens_front_radius: float | None = None,
    lens_front_asphericity: float | None = None,
    retina_radius: float | None = None,
    retina_asphericity: float | None = None,
    *,
    estimate_cornea_back: bool = False,
) -> GeometryType:
    """Create a geometry instance from clinically used parameters.

    All parameters are optional, and if not provided, the default values will be used.
    Sizes are specified in mm. If `estimate_cornea_back` is True, the back cornea radius will be estimated from the
    front cornea radius as `cornea_back_radius = 0.81 * cornea_front_radius`.

    Parameters
    ----------
    base : type[GeometryType]
        The base geometry class to use. Must be a subclass of EyeGeometry.
    axial_length : float, optional
        Axial length of the eye, measured from cornea front to retina.
    cornea_thickness : float, optional
        Thickness of the cornea.
    cornea_front_radius : float, optional
        Radius of curvature of the frontal cornea surface.
    cornea_front_asphericity : float, optional
        Asphericity of the frontal cornea surface.
    cornea_back_radius : float, optional
        Radius of curvature of the back cornea surface.
    cornea_back_asphericity : float, optional
        Asphericity of the back cornea surface.
    anterior_chamber_depth : float, optional
        Depth of the anterior chamber.
    pupil_radius : float, optional
        Radius of the pupil.
    lens_thickness : float, optional
        Thickness of the crystalline lens.
    lens_back_radius : float, optional
        Radius of curvature of the back lens surface.
    lens_back_asphericity : float, optional
        Asphericity of the back lens surface.
    lens_front_radius : float, optional
        Radius of curvature of the frontal lens surface.
    lens_front_asphericity : float, optional
        Asphericity of the frontal lens surface.
    retina_radius : float, optional
        Radius of curvature of the retina.
    retina_asphericity : float, optional
        Asphericity of the retina.
    estimate_cornea_back : bool, optional
        If True, the back cornea radius will be estimated from the front cornea radius. Default is `False`.

    Returns
    -------
    EyeGeometry
        An instance of the EyeGeometry class.

    Raises
    ------
    ValueError
        If the base geometry is not a class or if it is not a subclass of EyeGeometry.
    """
    if not isinstance(base, type):
        raise TypeError("The base geometry must be a class. Did you put parentheses after the class name?")

    if not issubclass(base, EyeGeometry):
        raise TypeError("The base geometry must be a subclass of EyeGeometry.")

    if estimate_cornea_back and cornea_back_radius is not None:
        warn("The cornea back radius was provided, but it will be ignored because estimate_cornea_back is True.")

    geometry = base()

    _update_attribute_if_specified(geometry.cornea_front, "thickness", cornea_thickness)
    _update_attribute_if_specified(geometry.cornea_front, "radius", cornea_front_radius)
    _update_attribute_if_specified(geometry.cornea_front, "asphericity", cornea_front_asphericity)

    if estimate_cornea_back:
        cornea_back_radius = 0.81 * geometry.cornea_front.radius

    _update_attribute_if_specified(geometry.cornea_back, "thickness", anterior_chamber_depth)
    _update_attribute_if_specified(geometry.cornea_back, "radius", cornea_back_radius)
    _update_attribute_if_specified(geometry.cornea_back, "asphericity", cornea_back_asphericity)

    _update_attribute_if_specified(geometry.pupil, "semi_diameter", pupil_radius)

    _update_attribute_if_specified(geometry.lens_front, "thickness", lens_thickness)
    _update_attribute_if_specified(geometry.lens_front, "radius", lens_front_radius)
    _update_attribute_if_specified(geometry.lens_front, "asphericity", lens_front_asphericity)

    vitreous_thickness = _calculate_vitreous_thickness(
        geometry, axial_length, cornea_thickness, anterior_chamber_depth, lens_thickness
    )

    if vitreous_thickness <= 0:
        raise ValueError(
            "The sum of the cornea thickness, anterior chamber depth and lens thickness is greater than "
            "or equal to the axial length."
        )

    _update_attribute_if_specified(geometry.lens_back, "thickness", vitreous_thickness)
    _update_attribute_if_specified(geometry.lens_back, "radius", lens_back_radius)
    _update_attribute_if_specified(geometry.lens_back, "asphericity", lens_back_asphericity)

    _update_attribute_if_specified(geometry.retina, "radius", retina_radius)
    _update_attribute_if_specified(geometry.retina, "asphericity", retina_asphericity)

    return geometry



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from visisipy.analysis.base import analysis
from visisipy.backend import get_backend

if TYPE_CHECKING:
    from visisipy import EyeModel
    from visisipy.refraction import FourierPowerVectorRefraction


__all__ = ("refraction",)


@analysis
def refraction(
    model: EyeModel | None,  # noqa: ARG001
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    pupil_diameter: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    *,
    use_higher_order_aberrations: bool = True,
    return_raw_result: bool = False,  # noqa: ARG001
) -> FourierPowerVectorRefraction | tuple[FourierPowerVectorRefraction, Any]:
    """Calculates the ocular refraction.

    The ocular refraction is calculated from Zernike standard coefficients and represented in Fourier power vector form.

    Parameters
    ----------
    model : EyeModel, optional
        The eye model to use for the refraction calculation. When `None`, the currently built model is used.
        Defaults to `None`.
    use_higher_order_aberrations : bool, optional
        If `True`, higher-order aberrations are used in the calculation. Defaults to `True`.
    field_coordinate : tuple[float, float], optional
        The field coordinate for the Zernike calculation. When `None`, the default field configured in the backend is
        used. Defaults to `None`.
    wavelength : float, optional
        The wavelength for the Zernike calculation. When `None`, the default wavelength configured in the backend is
        used. Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    pupil_diameter : float, optional
        The diameter of the pupil for the refraction calculation. Defaults to the pupil diameter configured in the
        model.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.

    Returns
    -------
     FourierPowerVectorRefraction
          The ocular refraction in Fourier power vector form.
    Any
        The raw analysis result from the backend.
    """
    return get_backend().analysis.refraction(
        use_higher_order_aberrations=use_higher_order_aberrations,
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        pupil_diameter=pupil_diameter,
        field_type=field_type,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import zospy as zp

from visisipy.backend import BaseBackend, _classproperty
from visisipy.opticstudio.analysis import OpticStudioAnalysis
from visisipy.opticstudio.models import BaseOpticStudioEye, OpticStudioEye

if TYPE_CHECKING:
    from collections.abc import Iterable
    from os import PathLike

    from zospy.zpcore import ZOS, OpticStudioSystem

    from visisipy import EyeModel


def initialize_opticstudio(
    mode: Literal["standalone", "extension"] = "standalone",
    zosapi_nethelper: str | None = None,
    opticstudio_directory: str | None = None,
) -> tuple[ZOS, OpticStudioSystem]:
    zos = zp.ZOS(zosapi_nethelper=zosapi_nethelper, opticstudio_directory=opticstudio_directory)
    oss = zos.connect(mode)

    return zos, oss


def _set_field_type(oss: OpticStudioSystem, field_type: str) -> None:
    if field_type == "angle":
        oss.SystemData.Fields.SetFieldType(zp.constants.SystemData.FieldType.Angle)
    elif field_type == "object_height":
        oss.SystemData.Fields.SetFieldType(zp.constants.SystemData.FieldType.ObjectHeight)
    else:
        raise ValueError("field_type must be either 'angle' or 'object_height'.")


def _set_fields(
    oss: OpticStudioSystem,
    coordinates: Iterable[tuple[float, float]],
) -> None:
    oss.SystemData.Fields.DeleteAllFields()

    for i, c in enumerate(coordinates):
        if i == 0:
            field = oss.SystemData.Fields.GetField(1)
            field.X, field.Y, field.Weight = c[0], c[1], 1
        else:
            oss.SystemData.Fields.AddField(c[0], c[1], 1)


def _remove_wavelenghts(oss: OpticStudioSystem) -> None:
    while oss.SystemData.Wavelengths.NumberOfWavelengths > 0:
        oss.SystemData.Wavelengths.RemoveWavelength(1)


class OpticStudioBackend(BaseBackend):
    zos: ZOS | None = None
    oss: OpticStudioSystem | None = None
    model: BaseOpticStudioEye | None = None

    @_classproperty
    def analysis(cls) -> OpticStudioAnalysis:  # noqa: N805
        """
        Provides access to the `OpticStudioAnalysis` instance.

        This property provides access to the `OpticStudioAnalysis` instance for performing various analyses on the optical
        system.

        Raises
        ------
        RuntimeError
            If the OpticStudio backend has not been initialized.

        Returns
        -------
        OpticStudioAnalysis
            The `OpticStudioAnalysis` instance.
        """
        if cls.oss is None:
            raise RuntimeError("The opticstudio backend has not been initialized.")

        return OpticStudioAnalysis(cls)

    @classmethod
    def initialize(
        cls,
        mode: Literal["standalone", "extension"] = "standalone",
        zosapi_nethelper: str | None = None,
        opticstudio_directory: str | None = None,
        ray_aiming: Literal["off", "paraxial", "real"] = "off",
    ) -> None:
        """
        Initializes the OpticStudio backend.

        This method connects to the OpticStudio backend and initializes a new optical system.

        Parameters
        ----------
        mode : Literal["standalone", "extension"], optional
            The mode to use when connecting to the OpticStudio backend. Can be either "standalone" or "extension".
            Defaults to "standalone".
        zosapi_nethelper : str | None, optional
            The path to the ZOS-API NetHelper DLL. If None, the path is determined automatically.
        opticstudio_directory : str | None, optional
            The path to the OpticStudio installation directory. If None, the path is determined automatically.
        ray_aiming : Literal["off", "paraxial", "real"], optional
            The ray aiming method to use. Can be either "off", "paraxial", or "real". Defaults to "off".
        """
        cls.zos, cls.oss = initialize_opticstudio(
            mode=mode,
            zosapi_nethelper=zosapi_nethelper,
            opticstudio_directory=opticstudio_directory,
        )

        cls.new_model(ray_aiming=ray_aiming)

    @classmethod
    def new_model(
        cls,
        *,
        save_old_model: bool = False,
        ray_aiming: Literal["off", "paraxial", "real"] = "off",
    ) -> None:
        """
        Initializes a new optical system model.

        This method initializes a new optical system model.
        """
        cls.oss.new(saveifneeded=save_old_model)

        if ray_aiming == "off":
            cls.oss.SystemData.RayAiming.RayAiming = zp.constants.SystemData.RayAimingMethod.Off
        elif ray_aiming == "paraxial":
            cls.oss.SystemData.RayAiming.RayAiming = zp.constants.SystemData.RayAimingMethod.Paraxial
        elif ray_aiming == "real":
            cls.oss.SystemData.RayAiming.RayAiming = zp.constants.SystemData.RayAimingMethod.Real
        else:
            raise ValueError("ray_aiming must be either 'off', 'paraxial', or 'real'.")

        cls.oss.SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.FloatByStopSize

    @classmethod
    def build_model(cls, model: EyeModel, *, replace_existing: bool = False, **kwargs) -> OpticStudioEye:
        """
        Builds an optical system based on the provided eye model.

        This method creates an OpticStudioEye instance from the provided eye model and builds the optical system.
        If `replace_existing` is True, any existing model is updated instead of building a completely new system.

        Parameters
        ----------
        model : EyeModel
            The eye model to be used for building the optical system model.
        replace_existing : bool, optional
            Whether to replace any existing model before building the new one. Defaults to False.
        **kwargs
            Additional keyword arguments to be passed to the OpticStudioEye build method.

        Returns
        -------
        OpticStudioEye
            The built optical system model.
        """
        if not replace_existing and cls.model is not None:
            cls.new_model()

        opticstudio_eye = OpticStudioEye(model)
        opticstudio_eye.build(cls.oss, replace_existing=replace_existing, **kwargs)

        cls.model = opticstudio_eye

        return opticstudio_eye

    @classmethod
    def clear_model(cls) -> None:
        """
        Clears the current optical system model.

        This method initializes a new optical system, discarding any existing model.
        """
        cls.oss.new(saveifneeded=False)
        cls.model = None

    @classmethod
    def save_model(cls, path: str | PathLike | None = None) -> None:
        """
        Saves the current optical system model.

        This method saves the current optical system model to the specified path. If no path is provided,
        it saves the model to the current working directory with the default name.

        Parameters
        ----------
        path : str | PathLike | None, optional
            The path where the model should be saved. If None, the model is saved in the current working directory.
        """
        if path is not None:
            cls.oss.save_as(path)
        else:
            cls.oss.save()

    @classmethod
    def disconnect(cls) -> None:
        """
        Disconnects the OpticStudio backend.

        This method closes the current optical system, sets the system and ZOS instances to None,
        and disconnects the ZOS instance.
        """
        cls.oss.close()
        cls.oss = None
        cls.zos.disconnect()
        cls.zos = None

    @classmethod
    def set_fields(
        cls,
        coordinates: Iterable[tuple[float, float]],
        field_type: Literal["angle", "object_height"] = "angle",
    ) -> None:
        """
        Sets the fields for the optical system.

        This method removes any existing fields and adds the new ones provided.

        Parameters
        ----------
        coordinates : Iterable[tuple[float, float]]
            An iterable of tuples representing the coordinates for the fields.
        field_type : Literal["angle", "object_height"], optional
            The type of field to be used in the optical system. Can be either "angle" or "object_height".
            Defaults to "angle".
        """
        _set_field_type(cls.oss, field_type)
        _set_fields(cls.oss, coordinates)

    @classmethod
    def set_wavelengths(cls, wavelengths: Iterable[float]) -> None:
        """
        Sets the wavelengths for the optical system.

        This method removes any existing wavelengths and adds the new ones provided.
        The weight for each wavelength is set to 1.0.

        Parameters
        ----------
        wavelengths : Iterable[float]
            An iterable of wavelengths to be set for the optical system.
        """
        _remove_wavelenghts(cls.oss)

        for w in wavelengths:
            cls.oss.SystemData.Wavelengths.AddWavelength(Wavelength=w, Weight=1.0)

    @classmethod
    def get_wavelength_number(cls, wavelength: float) -> int | None:
        """Returns the wavelength number for the given wavelength.

        If the wavelength is not found, `None` is returned.

        Parameters
        ----------
        wavelength: float
            The wavelength to find.

        Returns
        -------
        int | None
            The wavelength number, or `None` if the wavelength is not present.
        """
        for i in range(cls.oss.SystemData.Wavelengths.NumberOfWavelengths):
            if cls.oss.SystemData.Wavelengths.GetWavelength(i + 1).Wavelength == wavelength:
                return i + 1

        return None



from visisipy.analysis.cardinal_points import cardinal_points
from visisipy.analysis.raytracing import raytrace
from visisipy.analysis.refraction import refraction
from visisipy.analysis.zernike_standard_coefficients import (
    zernike_standard_coefficients,
)

__all__ = ("cardinal_points", "raytrace", "refraction", "zernike_standard_coefficients")



import visisipy
import visisipy.backend


class MockAnalysis:
    def cardinal_points(
        self,
        surface_1,
        surface_2,
    ):
        return None, None

    def raytrace(
        self,
        coordinates,
        wavelengths,
        field_type,
        pupil,
    ):
        return None, None

    def zernike_standard_coefficients(self, field_coordinate, wavelength, field_type, sampling, maximum_term):
        return None, None

    def refraction(
        self,
        use_higher_order_aberrations,
        field_coordinate,
        wavelength,
        pupil_diameter,
        field_type,
    ):
        return None, None


class MockBackend:
    analysis = MockAnalysis()


def test_cardinal_points_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.cardinal_points() is None
    assert visisipy.analysis.cardinal_points(return_raw_result=True) == (None, None)


def test_raytracing_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.raytrace(coordinates=[(0, 0)]) is None
    assert visisipy.analysis.raytrace(coordinates=[(0, 0)], return_raw_result=True) == (
        None,
        None,
    )


def test_zernike_standard_coefficients_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.zernike_standard_coefficients() is None
    assert visisipy.analysis.zernike_standard_coefficients(return_raw_result=True) == (
        None,
        None,
    )


def test_refraction_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.refraction() is None
    assert visisipy.analysis.refraction(return_raw_result=True) == (None, None)



from dataclasses import dataclass, field

__all__ = ("MaterialModel", "EyeMaterials", "NavarroMaterials")

from typing import Callable


@dataclass
class MaterialModel:
    """Optical properties of a material.

    Attributes
    ----------
    refractive_index : float
        The refractive index of the material.
    abbe_number : float
        The Abbe number (or V number) of the material, which is a measure of the material's dispersion
        (variation of refractive index with wavelength).
    partial_dispersion : float
        The partial dispersion of the material, which is the difference in the refractive index of the material
        at two specific wavelengths.
    """

    refractive_index: float
    abbe_number: float
    partial_dispersion: float


@dataclass
class EyeMaterials:
    """Material parameters of the elements of an eye.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.
    """

    cornea: MaterialModel
    aqueous: MaterialModel
    lens: MaterialModel
    vitreous: MaterialModel


def _material_model_factory(refractive_index, abbe_number, partial_dispersion) -> Callable[[], MaterialModel]:
    def factory() -> MaterialModel:
        return MaterialModel(refractive_index, abbe_number, partial_dispersion)

    return factory


@dataclass
class NavarroMaterials(EyeMaterials):
    """Material parameters of an eye, according to the Navarro model.

    The Navarro model defines refractive indices at various wavelengths.
    These data were used to fit the material models in OpticStudio.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.
    """

    cornea: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.37602751686,
            abbe_number=56.9362270454,
            partial_dispersion=0.0633737882164,
        )
    )
    aqueous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.33738703254,
            abbe_number=49.0704608205,
            partial_dispersion=0.0618839248407,
        )
    )
    lens: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.4200127433,
            abbe_number=48.0785554825,
            partial_dispersion=0.0838140446604,
        )
    )
    vitreous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.33602751687,
            abbe_number=50.8796247462,
            partial_dispersion=0.0531865765832,
        )
    )



import pytest

from visisipy import EyeGeometry, EyeMaterials, EyeModel
from visisipy.models.geometry import StandardSurface, Stop
from visisipy.models.materials import MaterialModel
from visisipy.opticstudio import OpticStudioEye

pytestmark = [pytest.mark.needs_opticstudio]


@pytest.fixture
def eye_model():
    geometry = EyeGeometry(
        cornea_front=StandardSurface(radius=7.72, asphericity=-0.26, thickness=0.55),
        cornea_back=StandardSurface(radius=6.50, asphericity=0, thickness=3.05),
        pupil=Stop(semi_diameter=1.348),
        lens_front=StandardSurface(radius=10.2, asphericity=-3.1316, thickness=4.0),
        lens_back=StandardSurface(radius=-6.0, asphericity=-1, thickness=16.3203),
        retina=StandardSurface(radius=-12.0, asphericity=0),
    )

    materials = EyeMaterials(
        cornea=MaterialModel(refractive_index=1.3777, abbe_number=0, partial_dispersion=0),
        aqueous=MaterialModel(refractive_index=1.3391, abbe_number=0, partial_dispersion=0),
        lens=MaterialModel(refractive_index=1.4222, abbe_number=0, partial_dispersion=0),
        vitreous=MaterialModel(refractive_index=1.3377, abbe_number=0, partial_dispersion=0),
    )

    return EyeModel(geometry=geometry, materials=materials)


class TestOpticStudioEye:
    def test_init(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)

        assert opticstudio_eye.eye_model == eye_model

    def test_build_cornea_front(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        assert opticstudio_eye.cornea_front.thickness == pytest.approx(eye_model.geometry.cornea_front.thickness)
        assert opticstudio_eye.cornea_front.radius == pytest.approx(eye_model.geometry.cornea_front.radius)
        assert opticstudio_eye.cornea_front.conic == pytest.approx(eye_model.geometry.cornea_front.asphericity)

    def test_build_cornea_back(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        assert opticstudio_eye.cornea_back.thickness == pytest.approx(eye_model.geometry.cornea_back.thickness)
        assert opticstudio_eye.cornea_back.radius == pytest.approx(eye_model.geometry.cornea_back.radius)
        assert opticstudio_eye.cornea_back.conic == pytest.approx(eye_model.geometry.cornea_back.asphericity)

    def test_build_pupil(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        assert opticstudio_eye.pupil.semi_diameter == pytest.approx(eye_model.geometry.pupil.semi_diameter)
        assert opticstudio_eye.pupil.is_stop

    def test_build_lens_front(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        assert opticstudio_eye.lens_front.thickness == pytest.approx(eye_model.geometry.lens_front.thickness)
        assert opticstudio_eye.lens_front.radius == pytest.approx(eye_model.geometry.lens_front.radius)
        assert opticstudio_eye.lens_front.conic == pytest.approx(eye_model.geometry.lens_front.asphericity)

    def test_build_lens_back(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        assert opticstudio_eye.lens_back.thickness == pytest.approx(eye_model.geometry.lens_back.thickness)
        assert opticstudio_eye.lens_back.radius == pytest.approx(eye_model.geometry.lens_back.radius)
        assert opticstudio_eye.lens_back.conic == pytest.approx(eye_model.geometry.lens_back.asphericity)

    def test_build_retina(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        assert opticstudio_eye.retina.radius == pytest.approx(eye_model.geometry.retina.radius)
        assert opticstudio_eye.retina.conic == pytest.approx(eye_model.geometry.retina.asphericity)
        assert opticstudio_eye.retina.surface.IsImage

    def test_build_pupil_not_stop_raises_valueerror(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.lens_front._is_stop = True

        with pytest.raises(ValueError, match="The pupil is not located at the stop position"):
            opticstudio_eye.build(new_oss)

    def test_build_retina_not_image_raises_valueerror(self, new_oss, eye_model):
        new_oss.LDE.InsertNewSurfaceAt(new_oss.LDE.NumberOfSurfaces)
        opticstudio_eye = OpticStudioEye(eye_model)

        with pytest.raises(ValueError, match="The retina is not located at the image position"):
            opticstudio_eye.build(new_oss)

    def test_set_cornea_front_property(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.cornea_front.comment = "new comment"

        assert new_oss.LDE.GetSurfaceAt(1).Comment == "new comment"

    def test_set_cornea_back_property(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.cornea_back.comment = "new comment"

        assert new_oss.LDE.GetSurfaceAt(2).Comment == "new comment"

    def test_set_pupil_property(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.pupil.comment = "new comment"

        assert new_oss.LDE.GetSurfaceAt(3).Comment == "new comment"

    def test_set_lens_front_property(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.lens_front.comment = "new comment"

        assert new_oss.LDE.GetSurfaceAt(4).Comment == "new comment"

    def test_set_lens_back_property(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.lens_back.comment = "new comment"

        assert new_oss.LDE.GetSurfaceAt(5).Comment == "new comment"

    def test_set_retina_property(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.retina.comment = "new comment"

        assert new_oss.LDE.GetSurfaceAt(6).Comment == "new comment"

    @pytest.mark.parametrize("build", [True, False])
    def test_surfaces(self, new_oss, eye_model, build):
        opticstudio_eye = OpticStudioEye(eye_model)

        if build:
            opticstudio_eye.build(new_oss)

        assert opticstudio_eye.surfaces == {
            "cornea_front": opticstudio_eye.cornea_front,
            "cornea_back": opticstudio_eye.cornea_back,
            "pupil": opticstudio_eye.pupil,
            "lens_front": opticstudio_eye.lens_front,
            "lens_back": opticstudio_eye.lens_back,
            "retina": opticstudio_eye.retina,
        }

    def test_update_surfaces(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.update_surfaces("comment", "new comment")

        assert opticstudio_eye.cornea_front.comment == "new comment"
        assert opticstudio_eye.cornea_back.comment == "new comment"
        assert opticstudio_eye.pupil.comment == "new comment"
        assert opticstudio_eye.lens_front.comment == "new comment"
        assert opticstudio_eye.lens_back.comment == "new comment"
        assert opticstudio_eye.retina.comment == "new comment"

    def test_update_surfaces_subset(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        opticstudio_eye.update_surfaces("comment", "new comment", surfaces=["cornea_front", "lens_back"])

        assert opticstudio_eye.cornea_front.comment == "new comment"
        assert opticstudio_eye.cornea_back.comment == "cornea back / aqueous"
        assert opticstudio_eye.pupil.comment == "pupil"
        assert opticstudio_eye.lens_front.comment == "lens front"
        assert opticstudio_eye.lens_back.comment == "new comment"
        assert opticstudio_eye.retina.comment == "retina"

    def test_relink_surfaces(self, new_oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(new_oss)

        new_oss.LDE.InsertNewSurfaceAt(1).Comment = "dummy"

        assert opticstudio_eye.relink_surfaces(new_oss)
        assert opticstudio_eye.cornea_front.surface.SurfaceNumber == 2
        assert opticstudio_eye.cornea_front.comment == "cornea front"
        assert opticstudio_eye.cornea_back.surface.SurfaceNumber == 3
        assert opticstudio_eye.cornea_back.comment == "cornea back / aqueous"
        assert opticstudio_eye.pupil.surface.SurfaceNumber == 4
        assert opticstudio_eye.pupil.comment == "pupil"
        assert opticstudio_eye.lens_front.surface.SurfaceNumber == 5
        assert opticstudio_eye.lens_front.comment == "lens front"
        assert opticstudio_eye.lens_back.surface.SurfaceNumber == 6
        assert opticstudio_eye.lens_back.comment == "lens back / vitreous"
        assert opticstudio_eye.retina.surface.SurfaceNumber == 7
        assert opticstudio_eye.retina.comment == "retina"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from visisipy.analysis.base import analysis
from visisipy.backend import get_backend

if TYPE_CHECKING:
    from visisipy.models import EyeModel


@analysis
def zernike_standard_coefficients(
    model: EyeModel | None,  # noqa: ARG001
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    sampling: str = "512x512",
    maximum_term: int = 45,
    *,
    return_raw_result: bool = False,  # noqa: ARG001
) -> tuple[dict, Any]:
    """
    Calculates the Zernike standard coefficients at the retina surface.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate for the Zernike calculation. When `None`, the first field in the backend is used.
        Defaults to `None`.
    wavelength : float | None, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in the backend is used.
        Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    sampling : str, optional
        The sampling for the Zernike calculation. Defaults to "512x512".
    maximum_term : int, optional
        The maximum term for the Zernike calculation. Defaults to 45.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.

    Returns
    -------
    AttrDict
        ZOSPy Zernike standard coefficients analysis output.
    """
    return get_backend().analysis.zernike_standard_coefficients(
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=sampling,
        maximum_term=maximum_term,
    )



from contextlib import nullcontext as does_not_raise

import pytest
import zospy as zp

from visisipy import EyeModel
from visisipy.opticstudio.backend import OpticStudioBackend

pytestmark = [pytest.mark.needs_opticstudio]


class TestOpticStudioBackend:
    def test_initialize_opticstudio(self, opticstudio_backend):
        assert opticstudio_backend.zos is not None
        assert opticstudio_backend.oss is not None

    @pytest.mark.parametrize(
        "ray_aiming,ray_aiming_constant,expectation",
        [
            ("off", "Off", does_not_raise()),
            ("paraxial", "Paraxial", does_not_raise()),
            ("real", "Real", does_not_raise()),
            (
                "invalid",
                None,
                pytest.raises(ValueError, match="ray_aiming must be either 'off', 'paraxial', or 'real'"),
            ),
        ],
    )
    def test_new_model(self, opticstudio_backend, ray_aiming, ray_aiming_constant, expectation):
        with expectation:
            OpticStudioBackend.new_model(save_old_model=False, ray_aiming=ray_aiming)

            assert opticstudio_backend.oss.SystemData.RayAiming.RayAiming == zp.constants.process_constant(
                zp.constants.SystemData.RayAimingMethod, ray_aiming_constant
            )
            assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 3

    def test_build_model(self, opticstudio_backend):
        model = EyeModel()

        opticstudio_backend.build_model(model)

        assert opticstudio_backend.model is not None
        assert opticstudio_backend.model.eye_model == model
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 7

    def test_clear_model(self, opticstudio_backend):
        model = EyeModel()

        opticstudio_backend.build_model(model)
        assert opticstudio_backend.model is not None
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 7

        opticstudio_backend.clear_model()

        assert opticstudio_backend.model is None
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 3

    def test_save_model(self, opticstudio_backend, tmp_path):
        model = EyeModel()

        opticstudio_backend.build_model(model)

        # Save the model to a temporary path
        model_path = tmp_path / "model.zmx"
        opticstudio_backend.save_model(model_path)

        assert model_path.exists()

        # Remove the model file
        model_path.unlink()

        assert not model_path.exists()

        # Save without specifying a path
        opticstudio_backend.save_model()

        assert model_path.exists()

    def test_disconnect(self, opticstudio_backend):
        opticstudio_backend.disconnect()

        assert opticstudio_backend.zos is None
        assert opticstudio_backend.oss is None

    @pytest.mark.parametrize(
        "coordinates,field_type,field_constant,expectation",
        [
            ([(10, 10)], "angle", "Angle", does_not_raise()),
            (
                [(0, 0), (0, 10), (-10, 0), (10, -10)],
                "angle",
                "Angle",
                does_not_raise(),
            ),
            (
                [(3.14, 4.15), (-12, 1)],
                "object_height",
                "ObjectHeight",
                does_not_raise(),
            ),
            (
                [(0, 0)],
                "invalid",
                None,
                pytest.raises(
                    ValueError,
                    match="field_type must be either 'angle' or 'object_height'",
                ),
            ),
        ],
    )
    def test_set_fields(self, opticstudio_backend, coordinates, field_type, field_constant, expectation):
        with expectation:
            opticstudio_backend.set_fields(coordinates, field_type)

            assert opticstudio_backend.oss.SystemData.Fields.NumberOfFields == len(coordinates)
            for i in range(len(coordinates)):
                field = opticstudio_backend.oss.SystemData.Fields.GetField(i + 1)
                assert coordinates[i] == (field.X, field.Y)

            assert opticstudio_backend.oss.SystemData.Fields.GetFieldType() == zp.constants.process_constant(
                zp.constants.SystemData.FieldType, field_constant
            )

    def test_set_wavelengths(self, opticstudio_backend):
        opticstudio_backend.set_wavelengths([0.543, 0.650])

        assert opticstudio_backend.oss.SystemData.Wavelengths.NumberOfWavelengths == 2
        assert opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(1).Wavelength == 0.543
        assert opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(2).Wavelength == 0.650

    def test_get_wavelength_number(self, opticstudio_backend):
        opticstudio_backend.set_wavelengths([0.543, 0.650])

        assert opticstudio_backend.get_wavelength_number(0.543) == 1
        assert opticstudio_backend.get_wavelength_number(0.650) == 2
        assert opticstudio_backend.get_wavelength_number(1.234) is None



from contextlib import nullcontext as does_not_raise

import pytest

from visisipy.models import EyeModel

pytestmark = [pytest.mark.needs_opticstudio]


@pytest.fixture
def opticstudio_analysis(opticstudio_backend):
    return opticstudio_backend.analysis


class TestCardinalPointsAnalysis:
    @pytest.mark.parametrize(
        "surface_1,surface_2,expectation",
        [
            (None, None, does_not_raise()),
            (1, 6, does_not_raise()),
            (2, 4, does_not_raise()),
            (
                -1,
                7,
                pytest.raises(
                    ValueError,
                    match="surface_1 and surface_2 must be between 1 and the number of surfaces in the system",
                ),
            ),
            (
                1,
                8,
                pytest.raises(
                    ValueError,
                    match="surface_1 and surface_2 must be between 1 and the number of surfaces in the system",
                ),
            ),
            (3, 3, pytest.raises(ValueError, match="surface_1 must be less than surface_2")),
            (4, 2, pytest.raises(ValueError, match="surface_1 must be less than surface_2")),
        ],
    )
    def test_cardinal_points(self, opticstudio_backend, opticstudio_analysis, surface_1, surface_2, expectation):
        opticstudio_backend.build_model(EyeModel())

        with expectation:
            assert opticstudio_analysis.cardinal_points(surface_1, surface_2)


class TestRayTraceAnalysis:
    def test_raytrace(self, opticstudio_analysis):
        coordinates = [(0, 0), (1, 1)]

        assert opticstudio_analysis.raytrace(coordinates)

    def test_raytrace_dataframe_structure(self, opticstudio_analysis):
        coordinates = [(0, 0), (1, 1)]

        result, _ = opticstudio_analysis.raytrace(coordinates)
        expected_columns = {
            "index",
            "x",
            "y",
            "z",
            "field",
            "wavelength",
            "surface",
            "comment",
        }

        assert set(result.columns) == expected_columns
        assert set(result.field.unique()) == set(coordinates)


class TestZernikeStandardCoefficientsAnalysis:
    @pytest.mark.parametrize("field_coordinate,wavelength", [(None, None), ((0, 0), 0.543), ((1, 1), 0.632)])
    def test_zernike_standard_coefficients(self, opticstudio_analysis, field_coordinate, wavelength):
        assert opticstudio_analysis.zernike_standard_coefficients(
            field_coordinate=field_coordinate, wavelength=wavelength
        )


class MockPupil:
    def __init__(self):
        self.semi_diameter = 1.0
        self.changed_semi_diameter = False

    @property
    def semi_diameter(self):
        return self._semi_diameter

    @semi_diameter.setter
    def semi_diameter(self, value):
        self.changed_semi_diameter = True
        self._semi_diameter = value


class MockOpticstudioModel:
    def __init__(self):
        self._pupil = MockPupil()

    @property
    def pupil(self):
        return self._pupil


class TestRefractionAnalysis:
    @pytest.mark.parametrize(
        "use_higher_order_aberrations,wavelength,pupil_diameter",
        [
            (False, None, None),
            (False, 0.543, 0.5),
            (True, 0.543, 1),
        ],
    )
    def test_refraction(
        self,
        opticstudio_analysis,
        use_higher_order_aberrations,
        wavelength,
        pupil_diameter,
        monkeypatch,
    ):
        monkeypatch.setattr(opticstudio_analysis._backend, "model", MockOpticstudioModel())

        assert opticstudio_analysis.refraction(
            use_higher_order_aberrations=use_higher_order_aberrations,
            wavelength=wavelength,
            pupil_diameter=pupil_diameter,
        )

    @pytest.mark.parametrize(
        "pupil_diameter,changed_pupil_diameter",
        [
            (None, False),
            (0.5, True),
        ],
    )
    def test_refraction_change_pupil(self, opticstudio_analysis, pupil_diameter, changed_pupil_diameter, monkeypatch):
        monkeypatch.setattr(opticstudio_analysis._backend, "model", MockOpticstudioModel())

        assert not opticstudio_analysis._backend.model.pupil.changed_semi_diameter

        opticstudio_analysis.refraction(pupil_diameter=pupil_diameter)

        assert opticstudio_analysis._backend.model.pupil.changed_semi_diameter == changed_pupil_diameter
        assert opticstudio_analysis._backend.model.pupil.semi_diameter == 1.0



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
import zospy as zp
from pandas import DataFrame

from visisipy.analysis.cardinal_points import CardinalPoints, CardinalPointsResult
from visisipy.backend import BaseAnalysis
from visisipy.refraction import FourierPowerVectorRefraction

if TYPE_CHECKING:
    from collections.abc import Iterable

    from zospy.api import _ZOSAPI
    from zospy.zpcore import OpticStudioSystem

    from visisipy.opticstudio.backend import OpticStudioBackend

__all__ = ("OpticStudioAnalysis",)


def _iter_fields(oss: OpticStudioSystem) -> tuple[int, _ZOSAPI.SystemData.IField]:
    for i in range(oss.SystemData.Fields.NumberOfFields):
        field = oss.SystemData.Fields.GetField(i + 1)
        yield field.FieldNumber, field


def _iter_wavelengths(oss: OpticStudioSystem) -> tuple[int, float]:
    for i in range(oss.SystemData.Wavelengths.NumberOfWavelengths):
        yield i + 1, oss.SystemData.Wavelengths.GetWavelength(i + 1)


def _build_cardinal_points_result(cardinal_points_result: zp.analyses.base.AttrDict) -> CardinalPointsResult:
    return CardinalPointsResult(
        focal_lengths=CardinalPoints(
            image=cardinal_points_result.Data["Image Space"]["Focal Length"],
            object=cardinal_points_result.Data["Object Space"]["Focal Length"],
        ),
        focal_points=CardinalPoints(
            image=cardinal_points_result.Data["Image Space"]["Focal Planes"],
            object=cardinal_points_result.Data["Object Space"]["Focal Planes"],
        ),
        principal_points=CardinalPoints(
            image=cardinal_points_result.Data["Image Space"]["Principal Planes"],
            object=cardinal_points_result.Data["Object Space"]["Principal Planes"],
        ),
        anti_principal_points=CardinalPoints(
            image=cardinal_points_result.Data["Image Space"]["Anti-Principal Planes"],
            object=cardinal_points_result.Data["Object Space"]["Anti-Principal Planes"],
        ),
        anti_nodal_points=CardinalPoints(
            image=cardinal_points_result.Data["Image Space"]["Anti-Nodal Planes"],
            object=cardinal_points_result.Data["Object Space"]["Anti-Nodal Planes"],
        ),
        nodal_points=CardinalPoints(
            image=cardinal_points_result.Data["Image Space"]["Nodal Planes"],
            object=cardinal_points_result.Data["Object Space"]["Nodal Planes"],
        ),
    )


def _build_raytrace_result(raytrace_results: list[DataFrame]) -> DataFrame:
    columns = {
        "Field": "field",
        "Wavelength": "wavelength",
        "Surf": "surface",
        "Comment": "comment",
        "X-coordinate": "x",
        "Y-coordinate": "y",
        "Z-coordinate": "z",
    }

    return pd.concat(raytrace_results)[columns.keys()].rename(columns=columns).reset_index()


def _get_zernike_coefficient(zernike_result: zp.analyses.base.AttrDict, coefficient: int) -> float:
    return zernike_result.Data.Coefficients.loc["Z" + str(coefficient)].Value


def _zernike_data_to_refraction(
    zernike_data: zp.analyses.base.AttrDict,
    pupil_data: zp.functions.lde.PupilData,
    wavelength: float,
    *,
    use_higher_order_aberrations: bool = True,
) -> FourierPowerVectorRefraction:
    z4 = _get_zernike_coefficient(zernike_data, 4) * wavelength * 4 * np.sqrt(3)
    z11 = _get_zernike_coefficient(zernike_data, 11) * wavelength * 12 * np.sqrt(5)
    z22 = _get_zernike_coefficient(zernike_data, 22) * wavelength * 24 * np.sqrt(7)
    z37 = _get_zernike_coefficient(zernike_data, 37) * wavelength * 40 * np.sqrt(9)

    z6 = _get_zernike_coefficient(zernike_data, 6) * wavelength * 2 * np.sqrt(6)
    z12 = _get_zernike_coefficient(zernike_data, 12) * wavelength * 6 * np.sqrt(10)
    z24 = _get_zernike_coefficient(zernike_data, 24) * wavelength * 12 * np.sqrt(14)
    z38 = _get_zernike_coefficient(zernike_data, 38) * wavelength * 60 * np.sqrt(2)

    z5 = _get_zernike_coefficient(zernike_data, 5) * wavelength * 2 * np.sqrt(6)
    z13 = _get_zernike_coefficient(zernike_data, 13) * wavelength * 6 * np.sqrt(10)
    z23 = _get_zernike_coefficient(zernike_data, 23) * wavelength * 12 * np.sqrt(14)
    z39 = _get_zernike_coefficient(zernike_data, 39) * wavelength * 60 * np.sqrt(2)

    exit_pupil_radius = pupil_data.ExitPupilDiameter / 2

    if use_higher_order_aberrations:
        return FourierPowerVectorRefraction(
            M=(-z4 + z11 - z22 + z37) / (exit_pupil_radius**2),
            J0=(-z6 + z12 - z24 + z38) / (exit_pupil_radius**2),
            J45=(-z5 + z13 - z23 + z39) / (exit_pupil_radius**2),
        )

    return FourierPowerVectorRefraction(
        M=(-z4) / (exit_pupil_radius**2),
        J0=(-z6) / (exit_pupil_radius**2),
        J45=(-z5) / (exit_pupil_radius**2),
    )


class OpticStudioAnalysis(BaseAnalysis):
    """
    Analyses for the OpticStudio backend.
    """

    def __init__(self, backend: OpticStudioBackend):
        self._backend = backend
        self._oss = backend.oss

    def cardinal_points(
        self, surface_1: int | None = None, surface_2: int | None = None
    ) -> tuple[CardinalPointsResult, zp.analyses.base.AnalysisResult]:
        """
        Get the cardinal points of the system between `surface_1` and `surface_2`.

        Parameters
        ----------
        surface_1 : int | None, optional
            The first surface to be used in the analysis. If `None`, the first surface in the system will be used.
            Defaults to `None`.
        surface_2 : int | None, optional
            The second surface to be used in the analysis. If `None`, the last surface in the system will be used.
            Defaults to `None`.

        Returns
        -------
        CardinalPointsResult
            The cardinal points of the system.

        Raises
        ------
        ValueError
            If `surface_1` or `surface_2` are not between 1 and the number of surfaces in the system, or if `surface_1`
            is greater than or equal to `surface_2`.
        """
        surface_1 = surface_1 or 1
        surface_2 = surface_2 or self._oss.LDE.NumberOfSurfaces - 1

        if surface_1 < 1 or surface_2 > self._oss.LDE.NumberOfSurfaces - 1:
            raise ValueError("surface_1 and surface_2 must be between 1 and the number of surfaces in the system.")

        if surface_1 >= surface_2:
            raise ValueError("surface_1 must be less than surface_2.")

        cardinal_points_result = zp.analyses.reports.cardinal_points(
            self._oss,
            surface_1=surface_1,
            surface_2=surface_2,
        )

        return _build_cardinal_points_result(cardinal_points_result), cardinal_points_result

    def raytrace(
        self,
        coordinates: Iterable[tuple[float, float]],
        wavelengths: Iterable[float] = (0.543,),
        field_type: Literal["angle", "object_height"] = "angle",
        pupil: tuple[float, float] = (0, 0),
    ) -> tuple[DataFrame, list[zp.analyses.base.AnalysisResult]]:
        """
        Perform a ray trace analysis using the given parameters.
        The ray trace is performed for each wavelength and field in the system, using the Single Ray Trace analysis
        in OpticStudio.

        The analysis returns a Dataframe with the following columns:

        - field: The field coordinates for the ray trace.
        - wavelength: The wavelength used in the ray trace.
        - surface: The surface number in the system.
        - comment: The comment for the surface.
        - x: The X-coordinate of the ray trace.
        - y: The Y-coordinate of the ray trace.
        - z: The Z-coordinate of the ray trace.

        Parameters
        ----------
        coordinates : Iterable[tuple[float, float]]
            An iterable of tuples representing the coordinates for the ray trace.
            If `field_type` is "angle", the coordinates should be the angles along the (X, Y) axes in degrees.
            If `field_type` is "object_height", the coordinates should be the object heights along the
            (X, Y) axes in mm.
        wavelengths : Iterable[float], optional
            An iterable of wavelengths to be used in the ray trace. Defaults to (0.543,).
        field_type : Literal["angle", "object_height"], optional
            The type of field to be used in the ray trace. Can be either "angle" or "object_height". Defaults to
            "angle".
        pupil : tuple[float, float], optional
            A tuple representing the pupil coordinates for the ray trace. Defaults to (0, 0).

        Returns
        -------
        DataFrame
            A pandas DataFrame containing the results of the ray trace analysis.
        """
        self._backend.set_fields(coordinates, field_type=field_type)
        self._backend.set_wavelengths(wavelengths)

        raytrace_results = []

        for wavelength_number, wavelength in _iter_wavelengths(self._backend.oss):
            for field_number, field in _iter_fields(self._backend.oss):
                raytrace_result = zp.analyses.raysandspots.single_ray_trace(
                    self._backend.oss,
                    px=pupil[0],
                    py=pupil[1],
                    field=field_number,
                    wavelength=wavelength_number,
                    global_coordinates=True,
                ).Data.RealRayTraceData

                raytrace_result.insert(0, "Field", [(field.X, field.Y)] * len(raytrace_result))
                raytrace_result.insert(0, "Wavelength", wavelength)

                raytrace_results.append(raytrace_result)

        return _build_raytrace_result(raytrace_results), raytrace_results

    def zernike_standard_coefficients(
        self,
        field_coordinate: tuple[float, float] | None = None,
        wavelength: float | None = None,
        field_type: Literal["angle", "object_height"] = "angle",
        sampling: str = "512x512",
        maximum_term: int = 45,
    ) -> tuple[zp.analyses.base.AttrDict, zp.analyses.base.AnalysisResult]:
        """
        Calculates the Zernike standard coefficients at the retina surface.

        Parameters
        ----------
        field_coordinate : tuple[float, float] | None, optional
            The field coordinate for the Zernike calculation. When `None`, the first field in OpticStudio is used.
            Defaults to `None`.
        wavelength : float | None, optional
            The wavelength for the Zernike calculation. When `None`, the first wavelength in OpticStudio is used.
            Defaults to `None`.
        field_type : Literal["angle", "object_height"], optional
            The type of field to be used when setting the field coordinate. This parameter is only used when
            `field_coordinate` is specified. Defaults to "angle".
        sampling : str, optional
            The sampling for the Zernike calculation. Defaults to "512x512".
        maximum_term : int, optional
            The maximum term for the Zernike calculation. Defaults to 45.

        Returns
        -------
        AttrDict
            ZOSPy Zernike standard coefficients analysis output.
        """
        wavelength_number = 1 if wavelength is None else self._backend.get_wavelength_number(wavelength)

        if wavelength_number is None:
            self._backend.set_wavelengths([wavelength])
            wavelength_number = 1

        if field_coordinate is not None:
            self._backend.set_fields([field_coordinate], field_type=field_type)

        zernike_result = zp.analyses.wavefront.zernike_standard_coefficients(
            self._oss,
            sampling=sampling,
            maximum_term=maximum_term,
            wavelength=wavelength_number,
            field=1,
            reference_opd_to_vertex=False,
            surface="Image",
            sx=0.0,
            sy=0.0,
            sr=1.0,
        )

        return zernike_result.Data, zernike_result

    def refraction(
        self,
        field_coordinate: tuple[float, float] | None = None,
        wavelength: float | None = None,
        pupil_diameter: float | None = None,
        field_type: Literal["angle", "object_height"] = "angle",
        *,
        use_higher_order_aberrations: bool = True,
    ) -> tuple[FourierPowerVectorRefraction, zp.analyses.base.AnalysisResult]:
        """Calculates the ocular refraction.

        The ocular refraction is calculated from Zernike standard coefficients and represented in Fourier power
        vector form.

        Parameters
        ----------
        use_higher_order_aberrations : bool, optional
            If `True`, higher-order aberrations are used in the calculation. Defaults to `True`.
        field_coordinate : tuple[float, float], optional
            The field coordinate for the Zernike calculation. When `None`, the first field in OpticStudio is used.
            Defaults to `None`.
        wavelength : float, optional
            The wavelength for the Zernike calculation. When `None`, the first wavelength in OpticStudio is used.
            Defaults to `None`.
        pupil_diameter : float, optional
            The diameter of the pupil for the refraction calculation. Defaults to the pupil diameter configured in the
            model.
        field_type : Literal["angle", "object_height"], optional
            The type of field to be used when setting the field coordinate. This parameter is only used when
            `field_coordinate` is specified. Defaults to "angle".

        Returns
        -------
         FourierPowerVectorRefraction
              The ocular refraction in Fourier power vector form.
        """
        # Get the wavelength from OpticStudio if not specified
        wavelength = self._oss.SystemData.Wavelengths.GetWavelength(1).Wavelength if wavelength is None else wavelength

        # Temporarily change the pupil diameter
        old_pupil_semi_diameter = None
        if pupil_diameter is not None:
            old_pupil_semi_diameter = self._backend.model.pupil.semi_diameter
            self._backend.model.pupil.semi_diameter = pupil_diameter / 2

        pupil_data = zp.functions.lde.get_pupil(self._oss)
        _, zernike_standard_coefficients = self.zernike_standard_coefficients(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
        )

        if old_pupil_semi_diameter is not None:
            self._backend.model.pupil.semi_diameter = old_pupil_semi_diameter

        return _zernike_data_to_refraction(
            zernike_standard_coefficients,
            pupil_data,
            wavelength,
            use_higher_order_aberrations=use_higher_order_aberrations,
        ), zernike_standard_coefficients
